%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% file: application_DoA.m
% author: Jakob Möderl (jakob.moederl@tugraz.at)
% institution: Graz University of Technology
% date: 29.11.2023
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Direction-of-Arrival (DOA) estimation example
%
% This file contains a demo example on how to use the fast variational
% block-sparse Bayesian learning algorithm for DOA estimation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear all;
close all;

addpath('fastBSBL/');
addpath('SBL_MF_matlab/');


%% Simulation Setup
SNRa_dB = 8;
d = 10;

% array spacing/wavelength
relative_spacing = 1/2;

% Number of antennas in the ULA
N_array = 100;
p_array = (0:N_array-1).';

% Steering vector of the array
steering_vector = @(theta) 1/sqrt(N_array) * exp(-2j*pi*relative_spacing*p_array*sin(theta(:).'));

% Setup grid of potential source DOAs
N_grid = 2*N_array;
xi_grid = (-N_grid:2:N_grid-1)/(N_grid);
theta_grid = asin(xi_grid);

% Generate Array Resposne Matrix based on the grid
Psi = steering_vector(theta_grid);


% Number of Sources
L = 3;    

% OSPA Distance Metrics
ospa_c = deg2rad(5);
ospa_p = 1;

% Here we could loop over several runs to calculate the mean OSPA over
% several realizations
for nSim = 1:1
    % coefficient for the first-order autoregressive source model
    beta = 0.95;
    % source amplitude covariance (over snapshots)
    Source_covar = toeplitz(beta.^(0:d-1));
    % Precision matrix of source amplitudes
    B = inv(Source_covar);
    % Generate source amplitudes according to AR-1 model
    L_i = chol(Source_covar)';    
    s = L_i * 1/sqrt(2) * (randn(d, L) + 1j*rand(d, L));
    
    % find indexs of active DOAs
    idx_s = [];
    for theta_true = [-2, 3,55]
      [minval, idx_min] = min(abs(theta_grid-deg2rad(theta_true)));
      idx_s = [idx_s, idx_min];
    end
    theta_true = theta_grid(idx_s);
    x_true = zeros(size(Psi, 2), d);
    x_true(idx_s, :) = s.';
    
    % Generate noiseless signal
    y_noiseless = Psi*x_true;
    
    % noise variance
    noise_samples = 1/sqrt(2) *  (randn(size(y_noiseless)) + 1j*randn(size(y_noiseless)));
    noise_power = mean(sum(abs(noise_samples).^2, 1));
    sig_power = mean(sum(abs(y_noiseless).^2, 1));
    noise_scaling = sqrt(N_array * sig_power/noise_power / (10^(SNRa_dB/10)));
    % save actual SNR for possible later evaluation
    SNR_dB(nSim) = 10*log10(sum(abs(y_noiseless(:)).^2)/sum(abs(noise_scaling*noise_samples(:).^2)));     
    % noisy signal
    y = y_noiseless + noise_scaling * noise_samples;
    
    
    %% BlockSBL
    % generate flattened dictionary matrix
    Phi = kron(Psi, eye(d));
    % vector of column indices such that we arrive at a block-sparse model
    y_hat = reshape(y.', [], 1);
    block_start = (1:d:size(Phi, 2)).';
    
    tic
    [x_hat, sigma_hat, lambda_hat, gamma_hat] = fastBSBL(y_hat, Phi, block_start, ...
      'rho', 1, 'prior_pdf_type', 'gamma', 'prior_b', 0, 'prior_c', 2, 'threshold', 1, 'calc_sigma_bar', 'full', 'B', B);
    time_fvSBL = toc;
    
    % reshape results to matrix form
    x_hat = reshape(x_hat, d, []).';
    % find active DOAs
    idx_active_SBL = find(~isinf(gamma_hat));
    theta_hat_FVSBL = theta_grid(idx_active_SBL);
    % calculate OSPA
    ospa_FVSBL(nSim) = rad2deg(ospa_dist(theta_true, theta_hat_FVSBL,ospa_c,ospa_p));
    
    
    
    %% DOA-SBL
    % reshape into a 3-dimensional tensor since the DOA-SBL can
    % (incoherently) run over several frequencies
    A = reshape(Psi, N_array, numel(theta_grid), 1);
    
    % DOA-SBL options
    options = SBLSet();
    options.gamma_range=10^(-3);
    options.fixedpoint = 1;
    options.convergence.error   = 10^(-4);
    options.status_report = 0;
    options.convergence.min_iteration = 1; 
    options.Nsource = 3;
    
    tic;
    [gamma_SBLGerstoft, SBLreport] = SBL_v4(A, y, options );
    time_PGSBL = toc;
    
    % threshold gamma based on the estimated noise variance
    th = SBLreport.results.final_iteration.noisepower * 10^(2/10);
    idx_theta_SBLGerstoft_1 = abs(gamma_SBLGerstoft) > th;
    theta_hat_SBLGerstoff_1 = theta_grid(idx_theta_SBLGerstoft_1);
    
    % oracle DOA-SBL using the 3 larges components
    [~, idx_theta_SBLGerstoft_2] = sort(gamma_SBLGerstoft, 'descend');
    theta_hat_SBLGerstoff_2 = theta_grid(idx_theta_SBLGerstoft_2(1:L));
    
    % calculate OSPA
    ospa_PGSBL_1(nSim) = rad2deg(ospa_dist(theta_true, theta_hat_SBLGerstoff_1,ospa_c,ospa_p));
    ospa_PGSBL_2(nSim) = rad2deg(ospa_dist(theta_true, theta_hat_SBLGerstoff_2,ospa_c,ospa_p));

    

    %% printing
    fprintf('run %3d/100: OSPA FV-BSBL: %1.4f    OSPA DOA-SBL: %1.4f     OSPA oracle DOA-SBL: %1.4f\n', nSim, ospa_FVSBL(nSim), ospa_PGSBL_1(nSim), ospa_PGSBL_2(nSim));



end
%% bartlett beamformer

OSPA = [ospa_FVSBL; ospa_PGSBL_1; ospa_PGSBL_2];
OSPA_mean = mean(OSPA, 2);
fprintf('-----------------------------------------------------------------------\n');
fprintf('Mean OSPA FV-SBL       Mean OSPA DOA-SBL       Mean OSPA oracle DOA-SBL\n')
fprintf('    %1.4f                   %1.4f                  %1.4f\n', ...
    OSPA_mean(1), OSPA_mean(2), OSPA_mean(3));
fprintf('-----------------------------------------------------------------------\n');

N_grid_bartlett = N_array;
xi_grid_bartlett = (-N_grid_bartlett:2:N_grid_bartlett-1)/(N_grid_bartlett);
theta_grid_bartlett = asin(xi_grid_bartlett);


Psi_bartlett = steering_vector(theta_grid_bartlett);

Pbartlett = mean(abs(Psi_bartlett' * y), 2);

fprintf('Plottign last run\n');

figure;
lgnd = {'Bartlett'};
hold on;
plot(rad2deg(theta_grid_bartlett), Pbartlett, 'k');

for idx_theta = 1:numel(theta_hat_FVSBL)
  plot(rad2deg(theta_hat_FVSBL(idx_theta)) *ones(2, 1), [0, max(Pbartlett)], 'r', 'linewidth', 2);
  lgnd{numel(lgnd)+1} = sprintf('blockSBL %d', idx_theta);
end

for idx_theta=1:numel(theta_hat_SBLGerstoff_2)
  plot(rad2deg(theta_hat_SBLGerstoff_2(idx_theta)) *ones(2, 1), [0, max(Pbartlett)], 'b:', 'linewidth', 2);
  lgnd{numel(lgnd)+1} = sprintf('SBLv4 %d', idx_theta);
end
for idx_theta = 1:numel(theta_true)
    plot(rad2deg(theta_true(idx_theta)) * ones(2, 1), [0, max(Pbartlett)], '--k', 'linewidth', 2)
end
lgnd{end+1} = 'True angles 1-3';
grid;
xlabel("angle $\theta$ in deg");
ylabel('Pbartlett');
legend(lgnd)

