function [x, Sigma, lambda, gamma, elbo] = fastBSBL(y, Phi, blk_start, NameValueArgs)%threshold, calc_sigma_bar, do_print)
% function FASTBSBL is the fast implementation of the block-sparse SBL
%   [x, Sigma, lambda, gamma] = fastBSBL(y, Phi, blk_start, ...)
%   estimates the block-sparse signal x, with blocks of size d(i), given
%   the design matrix Phi, noisy signal y and block start indices blk_start.
%   Additionally, the covariance Sigma, noise variance lambda and
%   prior variances gamma are estimated.
%
%
%   ======== INPUTS ========
%   y: (Nx1) measurement vector
%   Phi: (NxM) dictionary matrix
%   blk_start: (Kx1) integer vector with the indices of the first element of each block
%   Addtional Named arguments
%   rho: switch between complex and real-valued signal model.
%   prior_pdf_type: string, must be one of
%       {'jeffrey', 'gamma', inverse gamma'}
%   prior_a: double, prior parameter a>0 for the inverse gamma mixing pdf
%   prior_b: double, prior parameter b>0 for the gamma mixing pdf
%   prior_c: double, prior parameter c>\rho d_i for the gamma mixing pdf
%   threshold: (double, 0<threhsold<=1) pruning threshold. Default: 1
%   MaxIters: (double) Max numbers of iterations. Default: 30
%   do_print: (logical) print debug info in each iteration. Default: false
%   ======== OUTPUTS =========
%   x: (Mx1) vector of estimated weights
%   Sigma: estimate of the covariance of the nonzeros weights
%   lambda: (double) estimate of the noise precision
%   gamma: estimate of the prior covariances
%   elbo: value of the marginalized likelihood after each iteration
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% author: Jakob Möderl (jakob.moederl@tugraz.at)
% institution: Graz University of Technology
% date: 29.11.2023
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


arguments
  y (:, 1) {mustBeNumeric}
  Phi (:, :) {mustBeNumeric}
  blk_start {mustBeNumeric}
  NameValueArgs.B = 'eye';
  NameValueArgs.rho double = 0;
  NameValueArgs.prior_pdf_type {mustBeText} = 'jeffrey';
  NameValueArgs.prior_a double = 0;
  NameValueArgs.prior_b double = 0;
  NameValueArgs.prior_c double = 0;
  NameValueArgs.threshold double {mustBeLessThanOrEqual(NameValueArgs.threshold, 1)} = 1;
  NameValueArgs.calc_sigma_bar {mustBeText} = 'full';
  NameValueArgs.do_print logical = false;
  NameValueArgs.MaxIters double = 30;
end

calc_sigma_bar = NameValueArgs.calc_sigma_bar;
do_print = NameValueArgs.do_print;
threshold = NameValueArgs.threshold;
prior_pdf_type = NameValueArgs.prior_pdf_type;
prior_a = NameValueArgs.prior_a;
prior_b = NameValueArgs.prior_b;
prior_c = NameValueArgs.prior_c;
maxIters = NameValueArgs.MaxIters;

if any(NameValueArgs.rho == [1,1/2])
  rho = NameValueArgs.rho;
elseif NameValueArgs.rho == 0
  if isreal(y)
    rho = 1/2;
  else
    rho = 1;
  end
else
  error('parameter "rho" must either be 0 to set it automatically or either 1/2 or 1');
end

N = numel(y);
M = size(Phi, 2);
d = diff([blk_start; M+1]);

equal_block_sizes = all(d==d(1));

estimate_B = false;
if strcmpi(NameValueArgs.B, 'estimate')
  estimate_B = true;
  if equal_block_sizes
      B = eye(d(1));
  else
      error('Block sizes must be equal to estimate B');
  end
elseif strcmpi(NameValueArgs.B, 'eye')
    if equal_block_sizes
        B = eye(d(1));
    else
        for ii=1:numel(d)
            B = {};
            B{ii} = eye(d(ii));
        end
    end
elseif isnumeric(NameValueArgs.B) || iscell(NameValueArgs.B)
    B = NameValueArgs.B;
else
    error('B must be one of {"estiamte", "eye"}, a d x d matrix if all block sizes are equal or a cell array with matrices for each block')
end



%% initalizing
if strcmpi(prior_pdf_type, 'gamma')
    lambda = 2 * N/sum(abs(y).^2);
else
    lambda = N/sum(abs(y).^2);
end
Sigma = [];
x = zeros(M, 1);
gamma = inf * ones(M, 1);
% precompute Phi'*Phi for all blocks
PhiPhi = Phi'*Phi;

% precompute Phi' * y for all basis functions
Phiy = Phi'*y;

nn=1;
Sigma_update_counter = 0;
while true  
  if nn <= 3
      gamma_old = zeros(size(gamma));
  else
      gamma_old = gamma;
  end
  for ii=1:numel(blk_start)
    gamma(blk_start(ii):blk_start(ii)+d(ii)-1) = 0;
    idx_x_all = getIdxAll(gamma);
    idx_x_hat = getIdxHat(ii);

    if strcmpi(calc_sigma_bar, 'fast')
        if B ~= eye(d(1))
            error('Fast sigma update is not supported unless B=I');
        end
      idx_x_ii = getIdxII(ii);
      idx_x_bar = idx_x_all & ~idx_x_ii;
      %% Calc Ui, qi and si
      if isinf(gamma_old(blk_start(ii)))
        Sigma_bar_i = (lambda*PhiPhi(idx_x_ii, idx_x_ii) - lambda^2*PhiPhi(idx_x_ii, idx_x_bar)*Sigma*PhiPhi(idx_x_bar, idx_x_ii)) \ eye(d(ii));
        [Ui, S] = eig(ensureSymmetric(Sigma_bar_i));
        si = real(diag(S));
        Sigma_bar_Ei = [-lambda*Sigma_bar_i*PhiPhi(idx_x_ii, idx_x_bar)*Sigma, Sigma_bar_i];
        qi = lambda * Ui' * Sigma_bar_i * (Phiy(idx_x_ii) - lambda*PhiPhi(idx_x_ii, idx_x_bar)*Sigma*Phiy(idx_x_bar)); %Sigma_bar_Ei * [Phiy(idx_x_bar); Phiy(idx_x_ii)];
      else
        Sigma_bar = (lambda*PhiPhi(idx_x_all, idx_x_all) + diag(gamma(~isinf(gamma)))) \ eye(sum(idx_x_all));

        [Ui, S] = eig(ensureSymmetric(Sigma_bar(idx_x_hat, idx_x_hat)));
        si = real(diag(S));
        qi = lambda * Ui' * Sigma_bar(idx_x_hat, :) * Phiy(idx_x_all);
      end
    elseif strcmpi(calc_sigma_bar, 'full')
      B_bar = getBbar(gamma(blk_start));
      Li = chol(getBi(ii))';
      Sigma_bar = inv(lambda*PhiPhi(idx_x_all, idx_x_all) + B_bar);
      [Ui, S] = eig(ensureSymmetric(Li'*Sigma_bar(idx_x_hat, idx_x_hat)*Li));
      si = real(diag(S));
      qi = lambda * Ui' * Li' * Sigma_bar(idx_x_hat, :) * Phiy(idx_x_all);
    else
      error("Argument 'calc_sigma_bar' must be either 'full' or 'fast'");
    end
    % sanity check for numerical errors
    if any(real(si) < 0)
      warning('A nummerical Error occured');
    end
    
    %% calc gamma_i
    [A_poly, B_poly] = getPolynomialsAB(abs(qi).^2, si);
    if strcmpi(prior_pdf_type, 'jeffrey')
      gamma_i_ast = calcGamma_Jeffrey(A_poly, B_poly, threshold, gamma_old(blk_start(ii)));
    elseif strcmpi(prior_pdf_type, 'gamma')
      gamma_i_ast = calcGamma_Gamma(A_poly, B_poly, prior_a, prior_c, rho, threshold, gamma_old(blk_start(ii)));
    elseif strcmpi(prior_pdf_type, 'inverse gamma')
      gamma_i_ast = calcGamma_InverseGamma(A_poly, B_poly, prior_b, rho, threshold, gamma_old(blk_start(ii)));
    else
      error("Argument 'prior_pdf_type' must be one of {'gamma', 'inverse gamma', jeffrey'}");
    end
    gamma(blk_start(ii):blk_start(ii)+d(ii)-1) = gamma_i_ast;
%     if isempty(gamma_i_ast)
%       gamma(blk_start(ii):blk_start(ii)+d(ii)-1) = inf;
%     else
      

    %% Update sigma in case any gamma changed
    if strcmpi(calc_sigma_bar, 'fast')
      if ~(isinf(gamma(blk_start(ii))) && isinf(gamma_old(blk_start(ii))))
        idx_x_all = getIdxAll(gamma);
        gamma_vec = gamma(~isinf(gamma));
        Sigma = (lambda*PhiPhi(idx_x_all, idx_x_all) + diag(gamma_vec)) \ eye(numel(gamma_vec));
      end
    end
    
  end % end of inner loop over all blocks
  
  %% Final iteration steps
  % recalculate x_hat and Sigma_hat
  idx_x_all = getIdxAll(gamma);
  gamma_vec = gamma(~isinf(gamma));
  B_bar = getBbar(gamma(blk_start));
%   Sigma = ensureSymmetric(inv(lambda*PhiPhi(idx_x_all, idx_x_all) + diag(gamma_vec)));
  if strcmpi(calc_sigma_bar, 'full')
    Sigma = ensureSymmetric((lambda*PhiPhi(idx_x_all, idx_x_all) + B_bar)\eye(numel(gamma_vec)));
  end
  x(idx_x_all) = lambda*Sigma*Phiy(idx_x_all);
  x(~idx_x_all) = 0;
  est_erorr = sum(abs(y-Phi(:, idx_x_all)*x(idx_x_all)).^2);
  % calculate noise variance estimate
  lambda = N/real(trace(PhiPhi(idx_x_all, idx_x_all)*Sigma) + est_erorr);
  Sigma = ensureSymmetric((lambda*PhiPhi(idx_x_all, idx_x_all) + B_bar)\eye(numel(gamma_vec)));
      
  elbo(nn) = marginalizedLikelihoodFull();
  
  if numel(elbo) >= 2 && elbo(end) < elbo(end-1)
    warning('Elbo Decreased (%f --> %f) during iteration %d!', elbo(end-1), elbo(end), nn);
  end

  if do_print
    est_error_rel = est_erorr/sum(abs(y).^2);
    fprintf('Finished Iteration %d with %d groups active. Elbo: %f, Lambda:%f Approximation Error: %1.5f\n', nn, sum(~isinf(gamma)), elbo(nn), lambda, est_error_rel);
  end
  
  if sum(~isinf(gamma)) == 0
    warning('Did not find any components, aborting');
    break
  end
  
  if nn > 3 && (nn >= maxIters ...    % stop after maxIters iterations at maximum
      || (nn>=2 && all(~isinf(gamma)==~isinf(gamma_old)) ...     % check if the number of active groups changed
         && sum(abs(gamma(~isinf(gamma))-gamma_old(~isinf(gamma_old))).^2)/sum(gamma(~isinf(gamma)).^2) < 1e-3) ) % check how much the prior-variances changed from one iteration to the next
%     Sigma = ensureSymmetric(inv(lambda*PhiPhi(idx_x_all, idx_x_all) + diag(gamma_vec)));
    break;
  end
  
  % iteration counter
  nn=nn+1;
end % end main While loop
Niter = nn;
gamma = gamma(blk_start);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main function ends here
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Auxialry functions for index generation etc.

  function elbo_part = marginalizedLikelihoodGammaI(gamma_i)
    elbo_1 = log(gamma_i.*si./(1+gamma_i.*si));
    elbo_2 = gamma_i*abs(qi).^2./(1+gamma_i*si);
    elbo_part = sum(elbo_1) - sum(elbo_2);
  end

  function elbo = marginalizedLikelihoodFull()
    gamma_small = gamma(blk_start);
    elbo_1 = N*log(lambda);
    elbo_2 = sum(d(~isinf(gamma_small)) .* log(gamma_small(~isinf(gamma_small))));
    elbo_3 = logdet(Sigma);
    elbo_4 = -lambda*sum(abs(y).^2);
    elbo_5 = lambda^2*Phiy(idx_x_all)'*Sigma*Phiy(idx_x_all);
    elbo = real(elbo_1+elbo_2+elbo_3+elbo_4+elbo_5);
  end


  function idx=getIdxII(ii)
      idx = false(M, 1);
      idx(blk_start(ii):blk_start(ii)+d(ii)-1) = true;
  end

  function idx=getIdxHat(ii)
      idx_ii = getIdxII(ii);
      idx = idx_ii(~isinf(gamma));
  end

  function idx_all = getIdxAll(gamma_idx)
    idx_all = ~isinf(gamma_idx);
  end

  function B_bar = getBbar(gamma_idx)
    Bgamma = {};
    kk = 1;
    if equal_block_sizes
      for mm=1:numel(gamma_idx)
        if ~isinf(gamma_idx(mm))
          Bgamma{kk} = gamma_idx(mm)*B;
          kk = kk+1;
        end
      end
    else
        error('Using the intra-block correlation matrix B is currently not supported for unequal block sizes');
    end
    B_bar = blkdiag(Bgamma{:});
  end

  function Bi = getBi(mm)
    if equal_block_sizes
      Bi = B;
    else
      Bi = B{mm};
    end
  end
  


end



